"""Compute the excited states energies and MO coefficients using the STDDFT approach.

Index
-----
.. currentmodule:: nanoqm.workflows.workflow_stddft_spectrum
.. autosummary::
    workflow_stddft

"""

from __future__ import annotations

__all__ = ['workflow_stddft']

from os.path import join
from typing import Tuple, TYPE_CHECKING

import numpy as np
from scipy.linalg import sqrtm
from scipy.spatial.distance import cdist
from noodles import gather, schedule, unpack
from noodles.interface import PromisedObject
from qmflows.parsers import parse_string_xyz
from qmflows.type_hints import PathLike

from .. import logger
from ..common import (DictConfig, angs2au, change_mol_units, h2ev, hardness,
                      is_data_in_hdf5, number_spherical_functions_per_atom,
                      retrieve_hdf5_data, store_arrays_in_hdf5, xc)
from ..integrals.multipole_matrices import get_multipole_matrix
from ..schedule.components import calculate_mos
from .orbitals_type import select_orbitals_type

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from numpy import float64 as f8


def workflow_stddft(config: DictConfig) -> None:
    """Compute the excited states using simplified TDDFT.

    Both restricted and unrestricted orbitals calculations are available.

    Parameters
    ----------
    config
        Dictionary with the configuration to run the workflows

    """
    return select_orbitals_type(config, run_workflow_stddft)


def run_workflow_stddft(config: DictConfig) -> PromisedObject:
    """Compute the excited states using simplified TDDFT using `config`."""
    # Single Point calculations settings using CP2K
    mo_paths_hdf5, energy_paths_hdf5 = unpack(calculate_mos(config), 2)

    # Read structures
    molecules_au = [change_mol_units(parse_string_xyz(gs))
                    for i, gs in enumerate(config.geometries)
                    if (i % config.stride) == 0]

    # Noodles promised call
    scheduleTDDFT = schedule(compute_excited_states_tddft)

    results = gather(
        *[scheduleTDDFT(config, mo_paths_hdf5[i], DictConfig(
            {'i': i * config.stride, 'mol': mol}))
          for i, mol in enumerate(molecules_au)])

    return gather(results, energy_paths_hdf5)


def compute_excited_states_tddft(
        config: DictConfig, path_MOs: str, dict_input: DictConfig) -> None:
    """Compute the excited states properties (energy and coefficients).

    Take a given `mo_index_range`, the `tddft` method and `xc_dft` exchange functional.
    """
    logger.info("Reading energies and mo coefficients")
    # type of calculation
    energy, c_ao = retrieve_hdf5_data(config.path_hdf5, path_MOs)

    # Number of virtual orbitals
    nocc = config.active_space[0]
    nvirt = config.active_space[1]
    dict_input.update({"energy": energy, "c_ao": c_ao,
                       "nocc": nocc, "nvirt": nvirt})

    # Pass the molecule in Angstrom to the libint calculator
    copy_dict = DictConfig(dict_input.copy())
    copy_dict["mol"] = change_mol_units(dict_input["mol"], factor=1 / angs2au)

    # compute the multipoles if they are not stored
    multipoles = get_multipole_matrix(config, copy_dict, 'dipole')

    # read data from the HDF5 or calculate it on the fly
    dict_input["overlap"] = multipoles[0]

    # retrieve or compute the omega xia values
    omega, xia = get_omega_xia(config, dict_input)

    # add arrays to the dictionary
    dict_input.update(
        {"multipoles": multipoles[1:], "omega": omega, "xia": xia})

    compute_oscillator_strengths(config, dict_input)


def get_omega_xia(
        config: DictConfig, dict_input: DictConfig) -> Tuple[NDArray[f8], NDArray[f8]]:
    """Search for the multipole_matrices, Omega and xia values in the HDF5.

    if they are not available compute and store them.

    Returns
    -------
    tuple
        omega and xia numpy arrays.

    """
    tddft = config.tddft.lower()

    def compute_omega_xia():
        if tddft == 'sing_orb':
            return compute_sing_orb(dict_input)

        return compute_std_aproximation(config, dict_input)

    # search data in HDF5
    point = f'point_{dict_input.i + config.enumerate_from}'

    paths_omega_xia = [join(x, point) for x in ("omega", "xia")]

    if is_data_in_hdf5(config.path_hdf5, paths_omega_xia):
        ret = retrieve_hdf5_data(config.path_hdf5, paths_omega_xia)
        return ret[0], ret[1]

    else:
        omega, xia = compute_omega_xia()
        store_arrays_in_hdf5(
            config.path_hdf5, paths_omega_xia[0], omega, dtype=omega.dtype)
        store_arrays_in_hdf5(
            config.path_hdf5, paths_omega_xia[1], xia, dtype=xia.dtype)

        return omega, xia


def compute_sing_orb(inp: DictConfig) -> Tuple[NDArray[f8], NDArray[f8]]:
    """Compute the Single Orbital approximation."""
    energy, nocc, nvirt = tuple(inp[x]for x in ("energy", "nocc", "nvirt"))
    omega = -np.subtract(
        energy[:nocc].reshape(nocc, 1), energy[nocc:].reshape(nvirt, 1).T).reshape(nocc * nvirt)
    xia = np.eye(nocc * nvirt)

    return omega, xia


def compute_std_aproximation(
        config: DictConfig, dict_input: DictConfig) -> Tuple[NDArray[f8], NDArray[f8]]:
    """Compute the oscillator strenght using either the stda or stddft approximations."""
    logger.info("Reading or computing the dipole matrices")

    # Make a function tha returns in transition density charges
    logger.info("Computing the transition density charges")
    # multipoles[0] is the overlap matrix
    q = transition_density_charges(
        dict_input.mol, config, dict_input.overlap, dict_input.c_ao)

    # Make a function that compute the Mataga-Nishimoto-Ohno_Klopman
    # damped Columb and Excgange law functions
    logger.info(
        "Computing the gamma functions for Exchange and Coulomb integrals")
    gamma_J, gamma_K = compute_MNOK_integrals(dict_input["mol"], config.xc_dft)

    # Compute the Couloumb and Exchange integrals
    # If xc_dft is a pure functional, ax=0, thus the pqrs_J ints are not needed
    # and can be set to 0
    logger.info("Computing the Exchange and Coulomb integrals")
    if (xc(config.xc_dft)['type'] == 'pure'):
        size = dict_input.energy.size
        pqrs_J = np.zeros((size, size, size, size))
    else:
        pqrs_J = np.tensordot(q, np.tensordot(
            q, gamma_J, axes=(0, 1)), axes=(0, 2))
    pqrs_K = np.tensordot(q, np.tensordot(
        q, gamma_K, axes=(0, 1)), axes=(0, 2))

    # Construct the Tamm-Dancoff matrix A for each pair of i->a transition
    logger.info("Constructing the A matrix for TDDFT calculation")
    a_mat = construct_A_matrix_tddft(
        pqrs_J, pqrs_K, dict_input.nocc, dict_input.nvirt, config.xc_dft, dict_input.energy)

    if config.tddft == 'stddft':
        logger.info('sTDDFT has not been implemented yet !')
        # Solve the eigenvalue problem = A * cis = omega * cis
    elif config.tddft == 'stda':
        logger.info(
            "This is a TDA calculation ! \n Solving the eigenvalue problem")
        omega, xia = np.linalg.eig(a_mat)
    else:
        msg = f"The {config.tddft} method has not been implemented"
        raise NotImplementedError(msg)

    return omega, xia


def compute_oscillator_strengths(config: DictConfig, inp: DictConfig) -> None:
    """Compute oscillator strengths.

    The formula can be rearranged like this:
    f_I = 2/3 * np.sqrt(2 * omega_I) * sum_ia ( np.sqrt(e_diff_ia) * xia * tdm_x) ** 2 + y^2 + z^2
    """
    tddft = config.tddft.lower()
    # 1) Get the inp.energy matrix i->a. Size: Inp.Nocc * Inp.Nvirt
    delta_ia = -np.subtract(
        inp.energy[:inp.nocc].reshape(inp.nocc, 1),
        inp.energy[inp.nocc:].reshape(inp.nvirt, 1).T).reshape(inp.nocc * inp.nvirt)

    def compute_transition_matrix(matrix):
        if tddft == 'sing_orb':
            tm = np.stack(
                [np.sum(
                 delta_ia / inp.omega[i] * inp.xia[:, i] * matrix)
                 for i in range(inp.nocc * inp.nvirt)])
        else:
            tm = np.stack(
                [np.sum(
                 np.sqrt(2 * delta_ia / inp.omega[i]) * inp.xia[:, i] * matrix)
                 for i in range(inp.nocc * inp.nvirt)])
        return tm

    # 2) Compute the transition dipole matrix TDM(i->a)
    # Call the function that computes transition dipole moments integrals
    logger.info("Reading or computing the transition dipole matrix")

    def compute_tdmatrix(k):
        return np.linalg.multi_dot(
            [inp.c_ao[:, :inp.nocc].T, inp.multipoles[k, :, :],
             inp.c_ao[:, inp.nocc:]]).reshape(inp.nocc * inp.nvirt)

    td_matrices = (compute_tdmatrix(k) for k in range(3))

    # 3) Compute the transition dipole moments for each excited state i->a. Size: n_exc_states
    d_x, d_y, d_z = tuple(
        compute_transition_matrix(m) for m in td_matrices)

    # 4) Compute the oscillator strength
    f = 2 / 3 * inp.omega * (d_x ** 2 + d_y ** 2 + d_z ** 2)

    # Write to output
    inp.update({"dipole": (d_x, d_y, d_z), "oscillator": f})
    write_output(config, inp)


def write_output(config: DictConfig, inp: DictConfig) -> None:
    """Write the results using numpy functionality."""
    output = write_output_tddft(inp)

    path_output = join(config.workdir,
                       f'output_{inp.i + config.enumerate_from}_{config.tddft}.txt')
    fmt = '{:^5s}{:^14s}{:^8s}{:^11s}{:^11s}{:^11s}{:^11s}{:<5s}{:^10s}{:<5s}{:^11s}{:^11s}'
    header = fmt.format(
        'state', 'energy', 'f', 't_dip_x', 't_dip_y', 't_dip_y', 'weight',
        'from', 'energy', 'to', 'energy', 'delta_E')
    np.savetxt(path_output, output,
               fmt='%5d %10.3f %10.5f %10.5f %10.5f %10.5f %10.5f %3d %10.3f %3d %10.3f %10.3f',
               header=header)


def ex_descriptor(omega, f, xia, n_lowest, c_ao, s, tdm, tqm, nocc, nvirt, mol, config):
    """TODO: ADD DOCUMENTATION."""
    # Reshape xia
    xia_I = xia.reshape(nocc, nvirt, nocc * nvirt)

    # Transform the transition density matrix into AO basis
    d0I_ao = np.stack(
        np.linalg.multi_dot(
            [c_ao[:, :nocc], xia_I[:, :, i], c_ao[:, nocc:].T]) for i in range(n_lowest))

    # Compute omega in excition analysis for the lowest n excitations
    om = get_omega(d0I_ao, s, n_lowest)

    # Compute the distribution of positions for the hole and electron
    xh, yh, zh = get_exciton_positions(d0I_ao, s, tdm, n_lowest, 'hole')
    xe, ye, ze = get_exciton_positions(d0I_ao, s, tdm, n_lowest, 'electron')

    # Compute the distribution of the square of position for the hole and electron
    x2h, y2h, z2h = get_exciton_positions(d0I_ao, s, tqm, n_lowest, 'hole')
    x2e, y2e, z2e = get_exciton_positions(d0I_ao, s, tqm, n_lowest, 'electron')

    # Compute the distribution of both hole and electron positions
    xhxe, yhye, zhze = get_exciton_positions(d0I_ao, s, tdm, n_lowest, 'both')

    # Compute Descriptors

    # Compute exciton size:
    d_exc = np.sqrt(
        ((x2h - 2 * xhxe + x2e) + (y2h - 2 * yhye + y2e) + (z2h - 2 * zhze + z2e)) / om)

    # Compute centroid electron_hole distance
    d_he = np.abs(((xe - xh) + (ye - yh) + (ze - zh)) / om)

    # Compute hole and electron size
    sigma_h = np.sqrt(
        (x2h / om - (xh / om) ** 2) + (y2h / om - (yh / om) ** 2) + (z2h / om - (zh / om) ** 2))
    sigma_e = np.sqrt(
        (x2e / om - (xe / om) ** 2) + (y2e / om - (ye / om) ** 2) + (z2e / om - (ze / om) ** 2))

    # Compute Pearson coefficients
    cov = (xhxe - xh * xe) + (yhye - yh * ye) + (zhze - zh * ze)
    r_eh = cov / (sigma_h * sigma_e)

    # Compute approximate d_exc and binding energy
    omega_ab = get_omega_ab(d0I_ao, s, n_lowest, mol, config)
    r_ab = get_r_ab(mol)

    d_exc_apprx = np.stack(
        np.sqrt(np.sum(omega_ab[i, :, :] * (r_ab ** 2)) / om[i]) for i in range(n_lowest))
    # binding energy approximated
    xs = np.stack((omega_ab[i, :, :] / r_ab) for i in range(n_lowest))
    xs[np.isinf(xs)] = 0
    binding_en_apprx = np.stack(
        (np.sum(xs[i, :, :]) / om[i]) for i in range(n_lowest))

    descriptors = write_output_descriptors(
        d_exc, d_exc_apprx, d_he, sigma_h, sigma_e, r_eh, binding_en_apprx, n_lowest, omega, f)

    return descriptors


def write_output_descriptors(
        d_exc, d_exc_apprx, d_he, sigma_h, sigma_e, r_eh, binding_ex_apprx, n_lowest, omega, f):
    """TODO: add Documentation."""
    au2ang = 0.529177249
    ex_output = np.empty((n_lowest, 10))
    ex_output[:, 0] = np.arange(n_lowest) + 1
    ex_output[:, 1] = d_exc * au2ang
    ex_output[:, 2] = d_exc_apprx * au2ang
    ex_output[:, 3] = d_he * au2ang
    ex_output[:, 4] = sigma_h * au2ang
    ex_output[:, 5] = sigma_e * au2ang
    ex_output[:, 6] = r_eh * au2ang
    ex_output[:, 7] = binding_ex_apprx * 27.211 / 2  # in eV
    ex_output[:, 8] = omega[:n_lowest] * 27.211
    ex_output[:, 9] = f[:n_lowest]

    return ex_output


def get_omega(d0I_ao: NDArray[f8], s, n_lowest: int) -> NDArray[f8]:
    """TODO: add Documentation."""
    return np.stack([
        np.trace(np.linalg.multi_dot([d0I_ao[i, :, :].T, s, d0I_ao[i, :, :], s]))
        for i in range(n_lowest)
    ])


def get_r_ab(mol):
    """TODO: add Documentation."""
    coords = np.asarray([atom[1] for atom in mol])
    # Distance matrix between atoms A and B
    r_ab = cdist(coords, coords)
    return r_ab


def get_omega_ab(d0I_ao, s, n_lowest, mol, config):
    """TODO: add Documentation."""
    # Lowdin transformation of the transition density matrix
    n_atoms = len(mol)
    s_sqrt = sqrtm(s)
    d0I_mo = np.stack(
        np.linalg.multi_dot([s_sqrt, d0I_ao[i, :, :], s_sqrt]) for i in range(n_lowest))

    # Compute the number of spherical functions for each atom
    n_sph_atoms = number_spherical_functions_per_atom(
        mol, config['package_name'], config['basis_name'], config['path_hdf5'])

    # Compute omega_ab
    omega_ab = np.zeros((n_lowest, n_atoms, n_atoms))
    for i in range(n_lowest):
        index_a = 0
        for a in range(n_atoms):
            index_b = 0
            for b in range(n_atoms):
                omega_ab[i, a, b] = np.sum(
                    d0I_mo[i, index_a:(index_a + n_sph_atoms[a]),
                           index_b:(index_b + n_sph_atoms[b])] ** 2)
                index_b += n_sph_atoms[b]
            index_a += n_sph_atoms[a]

    return omega_ab


def get_exciton_positions(d0I_ao, s, moment, n_lowest, carrier):
    """TODO: add Documentation."""
    def compute_component_hole(k):
        return np.stack(
            np.trace(
                np.linalg.multi_dot([d0I_ao[i, :, :].T, moment[k, :, :], d0I_ao[i, :, :], s]))
            for i in range(n_lowest))

    def compute_component_electron(k):
        return np.stack(
            np.trace(
                np.linalg.multi_dot([d0I_ao[i, :, :].T, s, d0I_ao[i, :, :], moment[k, :, :]]))
            for i in range(n_lowest))

    def compute_component_he(k):
        return np.stack(
            np.trace(
                np.linalg.multi_dot(
                    [d0I_ao[i, :, :].T, moment[0, :, :], d0I_ao[i, :, :], moment[0, :, :]]))
            for i in range(n_lowest))

    if carrier == 'hole':
        return tuple(compute_component_hole(k) for k in range(3))
    elif carrier == 'electron':
        return tuple(compute_component_electron(k) for k in range(3))
    elif carrier == 'both':
        return tuple(compute_component_he(k) for k in range(3))
    else:
        raise RuntimeError(f"unkown option: {carrier}")


def write_output_tddft(inp: DictConfig) -> np.ndarray:
    """Write out as a table in plane text."""
    energy = inp.energy

    excs = [(i, a) for i in range(inp.nocc)
            for a in range(inp.nocc, inp.nvirt + inp.nocc)]

    output = np.empty((inp.nocc * inp.nvirt, 12))
    output[:, 0] = 0  # State number: we update it after reorder
    output[:, 1] = inp.omega * h2ev  # State energy in eV
    output[:, 2] = inp.oscillator  # Oscillator strength

    d_x, d_y, d_z = inp.dipole
    output[:, 3] = d_x  # Transition dipole moment in the x direction
    output[:, 4] = d_y  # Transition dipole moment in the y direction
    output[:, 5] = d_z  # Transition dipole moment in the z direction
    # Weight of the most important excitation
    output[:, 6] = np.hstack([np.max(inp.xia[:, i] ** 2)
                              for i in range(inp.nocc * inp.nvirt)])

    # Find the index of this transition
    index_weight = np.hstack([
        np.where(
            inp.xia[:, i] ** 2 == np.max(
                inp.xia[:, i] ** 2))
        for i in range(inp.nocc * inp.nvirt)]).reshape(inp.nocc * inp.nvirt)

    # Index of the hole for the most important excitation
    output[:, 7] = np.stack([excs[index_weight[i]][0]
                             for i in range(inp.nocc * inp.nvirt)]) + 1
    # These are the energies of the hole for the transition with the larger weight
    output[:, 8] = energy[output[:, 7].astype(int) - 1] * h2ev
    # Index of the electron for the most important excitation
    output[:, 9] = np.stack([excs[index_weight[i]][1]
                             for i in range(inp.nocc * inp.nvirt)]) + 1
    # These are the energies of the electron for the transition with the larger weight
    output[:, 10] = energy[output[:, 9].astype(int) - 1] * h2ev
    # This is the energy for the transition with the larger weight
    output[:, 11] = (
        energy[output[:, 9].astype(int) - 1] - energy[output[:, 7].astype(int) - 1]) * h2ev

    # Reorder the output in ascending order of energy
    output = output[output[:, 1].argsort()]
    # Give a state number in the correct order
    output[:, 0] = np.arange(inp.nocc * inp.nvirt) + 1

    return output


def transition_density_charges(mol, config, s, c_ao):
    """TODO: add Documentation."""
    n_atoms = len(mol)
    sqrt_s = sqrtm(s)
    c_mo = np.dot(sqrt_s, c_ao)
    # Size of the transition density tensor : n_atoms x n_mos x n_mos
    q = np.zeros((n_atoms, c_mo.shape[1], c_mo.shape[1]))
    n_sph_atoms = number_spherical_functions_per_atom(
        mol, config['package_name'], config.cp2k_general_settings['basis'], config['path_hdf5'])

    index = 0
    for i in range(n_atoms):
        q[i, :, :] = np.dot(
            c_mo[index:(index + n_sph_atoms[i]), :].T, c_mo[index:(index + n_sph_atoms[i]), :])
        index += n_sph_atoms[i]

    return q


def compute_MNOK_integrals(mol, xc_dft):
    """TODO: add Documentation."""
    n_atoms = len(mol)
    r_ab = get_r_ab(mol)
    hardness_vec = np.stack([hardness(m[0]) for m in mol]).reshape(n_atoms, 1)
    hard = np.add(hardness_vec, hardness_vec.T)
    beta = xc(xc_dft)['beta1'] + xc(xc_dft)['ax'] * xc(xc_dft)['beta2']
    alpha = xc(xc_dft)['alpha1'] + xc(xc_dft)['ax'] * xc(xc_dft)['alpha2']
    if (xc(xc_dft)['type'] == 'pure'):
        gamma_J = np.zeros((n_atoms, n_atoms))
    else:
        gamma_J = np.power(
            1 / (np.power(r_ab, beta) + np.power((xc(xc_dft)['ax'] * hard), -beta)),
            1 / beta,
        )
    gamma_K = np.power(
        1 / (np.power(r_ab, alpha) + np.power(hard, -alpha)),
        1 / alpha,
    )

    return gamma_J, gamma_K


def construct_A_matrix_tddft(pqrs_J, pqrs_K, nocc, nvirt, xc_dft, e):
    """TODO: add Documentation."""
    # This is the exchange integral entering the A matrix.
    #  It is in the format (nocc, nvirt, nocc, nvirt)
    k_iajb = pqrs_K[:nocc, nocc:, :nocc, nocc:].reshape(
        nocc * nvirt, nocc * nvirt)

    # This is the Coulomb integral entering in the A matrix.
    # It is in the format: (nocc, nocc, nvirt, nvirt)
    k_ijab_tmp = pqrs_J[:nocc, :nocc, nocc:, nocc:]

    # To get the correct order in the A matrix, i.e. (nocc, nvirt, nocc, nvirt),
    # we have to swap axes
    k_ijab = np.swapaxes(k_ijab_tmp, axis1=1, axis2=2).reshape(
        nocc * nvirt, nocc * nvirt)

    # They are in the m x m format where m is the number of excitations = nocc * nvirt
    a_mat = 2 * k_iajb - k_ijab

    # Generate a vector with all possible ea - ei energy differences
    e_diff = -np.subtract(
        e[:nocc].reshape(nocc, 1), e[nocc:].reshape(nvirt, 1).T).reshape(nocc * nvirt)
    np.fill_diagonal(a_mat, np.diag(a_mat) + e_diff)

    return a_mat



"""The Nano-QMFlows logger."""

from __future__ import annotations

import os
import sys
import types
import logging
import contextlib
from typing import ClassVar

from qmflows.type_hints import PathLike

__all__ = ["logger", "stdout_handler", "EnableFileHandler"]

#: The Nano-QMFlows logger.
logger = logging.getLogger("nanoqm")
logger.setLevel(logging.DEBUG)

qmflows_logger = logging.getLogger("qmflows")
noodles_logger = logging.getLogger("noodles")
noodles_logger.setLevel(logging.WARNING)

#: The Nano-QMFlows stdout handler.
stdout_handler = logging.StreamHandler(stream=sys.stdout)
stdout_handler.setLevel(logging.DEBUG)
stdout_handler.setFormatter(logging.Formatter(
    fmt='[%(asctime)s] %(levelname)s: %(message)s',
    datefmt='%H:%M:%S',
))
logger.addHandler(stdout_handler)


class EnableFileHandler(contextlib.ContextDecorator):
    """Add a file handler to the noodles, qmflows and nanoqm loggers.

    Attributes
    ----------
    handler : logging.FileHandler
        The relevant titular handler.

    """

    __slots__ = ("handler",)

    LOGGERS: ClassVar = (logger, qmflows_logger, noodles_logger)

    def __init__(self, path: PathLike) -> None:
        """Initialize the context manager.

        Parameters
        ----------
        path : path-like object
            Path to the log file.

        """
        self.handler = logging.FileHandler(os.fsdecode(path))
        self.handler.setLevel(logging.DEBUG)
        self.handler.setFormatter(logging.Formatter(
            fmt='%(asctime)s---%(levelname)s\n%(message)s\n',
            datefmt='%H:%M:%S',
        ))

    def __enter__(self) -> None:
        """Add the file handler."""
        for logger in self.LOGGERS:
            if self.handler not in logger.handlers:
                logger.addHandler(self.handler)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: types.TracebackType | None,
    ) -> None:
        """Remove the file handler."""
        for logger in self.LOGGERS:
            if self.handler in logger.handlers:
                logger.removeHandler(self.handler)



"""Crystal Orbital Overlap Population calculation.

Index
-----
.. currentmodule:: nanoqm.workflows.workflow_coop
.. autosummary::
    workflow_crystal_orbital_overlap_population

"""

from __future__ import annotations

__all__ = ['workflow_crystal_orbital_overlap_population']

from typing import Tuple, TYPE_CHECKING

import numpy as np

from qmflows.parsers import readXYZ

from .. import logger
from ..common import (DictConfig, MolXYZ, h2ev,
                      number_spherical_functions_per_atom, retrieve_hdf5_data)
from ..integrals.multipole_matrices import compute_matrix_multipole
from .initialization import initialize
from .tools import compute_single_point_eigenvalues_coefficients

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from numpy import float64 as f8, int64 as i8


def workflow_crystal_orbital_overlap_population(config: DictConfig) -> NDArray[f8]:
    """Compute the Crystal Orbital Overlap Population."""
    # Dictionary containing the general information
    config.update(initialize(config))

    # Checking hdf5 for eigenvalues and coefficients. If not present, they are
    # computed.
    compute_single_point_eigenvalues_coefficients(config)

    # Logger info
    logger.info("Starting COOP calculation.")

    # Get eigenvalues and coefficients from hdf5
    atomic_orbitals, energies = get_eigenvalues_coefficients(config)

    # Converting the xyz-file to a mol-file
    mol = readXYZ(config.path_traj_xyz)

    # Computing the indices of the atomic orbitals of the two selected
    # elements, and the overlap matrix that contains only elements related to
    # the two elements
    el_1_orbital_ind, el_2_orbital_ind, overlap_reduced = compute_overlap_and_atomic_orbitals(
        mol, config)

    # Compute the crystal orbital overlap population between the two selected
    # elements
    coop = compute_coop(
        atomic_orbitals,
        overlap_reduced,
        el_1_orbital_ind,
        el_2_orbital_ind,
    )

    # Lastly, we save the COOP as a txt-file
    result_coop = print_coop(energies, coop)
    logger.info("COOP calculation completed.")

    return result_coop


def get_eigenvalues_coefficients(config: DictConfig) -> Tuple[NDArray[f8], NDArray[f8]]:
    """Retrieve eigenvalues and coefficients from hdf5 file."""
    # Define paths to eigenvalues and coefficients hdf5
    node_path_coefficients = 'coefficients/point_0/'
    node_path_eigenvalues = 'eigenvalues/point_0'

    # Retrieves eigenvalues and coefficients
    atomic_orbitals = retrieve_hdf5_data(config.path_hdf5, node_path_coefficients)
    energies = retrieve_hdf5_data(config.path_hdf5, node_path_eigenvalues)

    # Energies converted from Hartree to eV
    energies *= h2ev

    # Return atomic orbitals and energies
    return atomic_orbitals, energies


def compute_overlap_and_atomic_orbitals(
    mol: MolXYZ,
    config: DictConfig,
) -> Tuple[NDArray[i8], NDArray[i8], NDArray[f8]]:
    """Compute the indices of the atomic orbitals of the two selected elements.

    Computes the overlap matrix, containing only the elements related to those two elements.
    """
    # Computing the overlap-matrix S
    overlap = compute_matrix_multipole(mol, config, 'overlap')

    # Computing number of spherical orbitals per atom
    sphericals = number_spherical_functions_per_atom(
        mol,
        'cp2k',
        config["cp2k_general_settings"]["basis"],
        config["path_hdf5"])

    # Getting the indices for the two selected elements
    element_1 = config["coop_elements"][0]
    element_2 = config["coop_elements"][1]

    element_1_index = [i for i, s in enumerate(mol) if element_1.lower() in s]
    element_2_index = [i for i, s in enumerate(mol) if element_2.lower() in s]

    # Making a list of the indices of the atomic orbitals for each of the two
    # elements
    atom_indices = np.zeros(len(mol) + 1, dtype=np.int64)
    atom_indices[1:] = np.cumsum(sphericals)

    _el_1_orbital_ind = [
        np.arange(sphericals[i], dtype=np.int64) + atom_indices[i] for i in element_1_index
    ]
    el_1_orbital_ind = np.reshape(
        _el_1_orbital_ind,
        len(element_1_index) * sphericals[element_1_index[0]],
    )

    _el_2_orbital_ind = [
        np.arange(sphericals[i], dtype=np.int64) + atom_indices[i] for i in element_2_index
    ]
    el_2_orbital_ind = np.reshape(
        _el_2_orbital_ind,
        len(element_2_index) * sphericals[element_2_index[0]],
    )

    # Reduced overlap matrix, containing only the elements related to the
    # overlap between element_1 and element_2
    # First select all the rows that belong to element_1
    overlap_reduced = overlap[el_1_orbital_ind, :]
    # Then select from those rows the columns that belong to species element_2
    overlap_reduced = overlap_reduced[:, el_2_orbital_ind]

    # Return lists of indices of atomic orbitals, and the reduced overlap
    # matrix
    return el_1_orbital_ind, el_2_orbital_ind, overlap_reduced


def compute_coop(
    atomic_orbitals: NDArray[f8],
    overlap_reduced: NDArray[f8],
    el_1_orbital_ind: NDArray[i8],
    el_2_orbital_ind: NDArray[i8],
) -> NDArray[f8]:
    """Define the function that computes the crystal orbital overlap population.

    Applying it to each column of the coefficent matrix.
    """
    # Define a function to be applied to each column of the coefficient matrix
    def coop_func(column_of_coefficient_matrix: NDArray[f8]) -> NDArray[f8]:
        # Multiply each coefficient-product with the relevant overlap, and sum
        # everything
        return np.sum(
            np.tensordot(
                column_of_coefficient_matrix[el_1_orbital_ind],
                column_of_coefficient_matrix[el_2_orbital_ind],
                0) * overlap_reduced)

    # Call the function
    coop = np.apply_along_axis(
        coop_func,
        0,
        atomic_orbitals)

    # Return the calculated crystal orbital overlap population
    return coop


def print_coop(energies: NDArray[f8], coop: NDArray[f8]) -> NDArray[f8]:
    """Save the COOP in a txt-file."""
    result_coop = np.zeros((len(coop), 2))
    result_coop[:, 0], result_coop[:, 1] = energies, coop
    np.savetxt('COOP.txt', result_coop)

    return result_coop



"""Module containing physical constants and `NamedTuple`s to store molecular orbitals, shell, etc.

Index
-----
.. currentmodule:: nanoqm.common
.. autosummary::
    DictConfig
    change_mol_units
    getmass
    number_spherical_functions_per_atom
    retrieve_hdf5_data
    is_data_in_hdf5
    store_arrays_in_hdf5

API
---
.. autoclass:: DictConfig
.. autofunction:: is_data_in_hdf5
.. autofunction:: retrieve_hdf5_data
.. autofunction:: number_spherical_functions_per_atom
.. autofunction:: store_arrays_in_hdf5

"""

from __future__ import annotations

__all__ = ['DictConfig', 'Matrix', 'Tensor3D', 'Vector',
           'change_mol_units', 'getmass', 'h2ev', 'hardness',
           'number_spherical_functions_per_atom', 'retrieve_hdf5_data',
           'is_data_in_hdf5', 'store_arrays_in_hdf5', 'UniqueSafeLoader',
           'valence_electrons', 'aux_fit']

import os
import json
from itertools import chain
from pathlib import Path
from typing import (Any, Dict, Iterable, List, Mapping, NamedTuple, Tuple,
                    Sequence, overload, TypeVar, TYPE_CHECKING, Iterator)

import h5py
import mendeleev
import numpy as np
from scipy.constants import physical_constants
from qmflows.common import AtomXYZ
from qmflows.type_hints import PathLike
from scm.plams import Atom, Molecule

from qmflows.yaml_utils import UniqueSafeLoader
from . import __path__ as nanoqm_path

if TYPE_CHECKING:
    import numpy.typing as npt

_T = TypeVar("_T")


_path_valence_electrons = Path(nanoqm_path[0]) / "basis" / "valence_electrons.json"
_path_aux_fit = Path(nanoqm_path[0]) / "basis" / "aux_fit.json"

with open(_path_valence_electrons, 'r') as f1, open(_path_aux_fit, 'r') as f2:
    valence_electrons: "dict[str, int]" = json.load(f1)
    aux_fit: "dict[str, list[int]]" = json.load(f2)


class DictConfig(Dict[str, Any]):
    """Class to extend the Dict class with `.` dot notation."""

    def __getattr__(self, attr: str) -> Any:
        """Extract key using dot notation."""
        return self.get(attr)

    def __setattr__(self, key: str, value: Any) -> None:
        """Set value using dot notation."""
        self.__setitem__(key, value)

    def __deepcopy__(self, _: object) -> "DictConfig":
        """Deepcopy of the Settings object."""
        return DictConfig(self.copy())


class BasisFormats(NamedTuple):
    """NamedTuple that contains the name/value for the basis formats."""

    name: str
    value: list[npt.NDArray[np.int64]]


def concat(xss: Iterable[Iterable[_T]]) -> List[_T]:
    """Concatenate of all the elements of a list."""
    return list(chain(*xss))


# ================> Constants <================

# Prevent a TypeError whenever scipy is mocked
if isinstance(physical_constants['atomic unit of length'][0], float):
    #: Angstrom to a.u
    angs2au = 1e-10 / physical_constants['atomic unit of length'][0]
    #: from femtoseconds to au
    femtosec2au = 1e-15 / physical_constants['atomic unit of time'][0]
    #: hartrees to electronvolts
    h2ev = physical_constants['Hartree energy in eV'][0]
    #: conversion from rydberg to meV
    r2meV = 1e3 * physical_constants['Rydberg constant times hc in eV'][0]
    #: conversion from fs to cm-1
    fs_to_cm = 1e13 * physical_constants['hertz-inverse meter relationship'][0]
    #: conversion from fs to nm
    fs_to_nm = 299.79246
    #: planck constant in eV * fs
    hbar = 1e15 * physical_constants['Planck constant over 2 pi in eV s'][0]
else:
    angs2au = femtosec2au = h2ev = r2meV = fs_to_cm = fs_to_nm = hbar = 1.0

# type hints
MolXYZ = List[AtomXYZ]
Vector = np.ndarray
Matrix = np.ndarray
Tensor3D = np.ndarray


def getmass(s: str) -> int:
    """Get the atomic mass for a given element s."""
    element = mendeleev.element(s.capitalize())
    return element.mass_number


def hardness(s: str) -> float:
    """Get the element hardness."""
    d = {
        'h': 6.4299, 'he': 12.5449, 'li': 2.3746, 'be': 3.4968, 'b': 4.619, 'c': 5.7410,
        'n': 6.8624, 'o': 7.9854, 'f': 9.1065, 'ne': 10.2303, 'na': 2.4441, 'mg': 3.0146,
        'al': 3.5849, 'si': 4.1551, 'p': 4.7258, 's': 5.2960, 'cl': 5.8662, 'ar': 6.4366,
        'k': 2.3273, 'ca': 2.7587, 'sc': 2.8582, 'ti': 2.9578, 'v': 3.0573, 'cr': 3.1567,
        'mn': 3.2564, 'fe': 3.3559, 'co': 3.4556, 'ni': 3.555, 'cu': 3.6544, 'zn': 3.7542,
        'ga': 4.1855, 'ge': 4.6166, 'as': 5.0662, 'se': 5.4795, 'br': 5.9111, 'kr': 6.3418,
        'rb': 2.1204, 'sr': 2.5374, 'y': 2.6335, 'zr': 2.7297, 'nb': 2.8260, 'mo': 2.9221,
        'tc': 3.0184, 'ru': 3.1146, 'rh': 3.2107, 'pd': 3.3069, 'ag': 3.4032, 'cd': 3.4994,
        'in': 3.9164, 'sn': 4.3332, 'sb': 4.7501, 'te': 5.167, 'i': 5.5839, 'xe': 6.0009,
        'cs': 0.6829, 'ba': 0.9201, 'la': 1.1571, 'ce': 1.3943, 'pr': 1.6315, 'nd': 1.8686,
        'pm': 2.1056, 'sm': 2.3427, 'eu': 2.5798, 'gd': 2.8170, 'tb': 3.0540, 'dy': 3.2912,
        'ho': 3.5283, 'er': 3.7655, 'tm': 4.0026, 'yb': 4.2395, 'lu': 4.4766, 'hf': 4.7065,
        'ta': 4.9508, 'w': 5.1879, 're': 5.4256, 'os': 5.6619, 'ir': 5.900, 'pt': 6.1367,
        'au': 6.3741, 'hg': 6.6103, 'tl': 1.7043, 'pb': 1.9435, 'bi': 2.1785, 'po': 2.4158,
        'at': 2.6528, 'rn': 2.8899, 'fr': 0.9882, 'ra': 1.2819, 'ac': 1.3497, 'th': 1.4175,
        'pa': 1.9368, 'u': 2.2305, 'np': 2.5241, 'pu': 3.0436, 'am': 3.4169, 'cm': 3.4050,
        'bk': 3.9244, 'cf': 4.2181, 'es': 4.5116, 'fm': 4.8051, 'md': 5.0100, 'no': 5.3926,
        'lr': 5.4607}
    return d[s] / 27.211


def xc(s: str) -> Dict[str, Any]:
    """Return the exchange functional composition."""
    d = {
        'pbe': {
            'type': 'pure', 'alpha1': 1.42, 'alpha2': 0.48, 'ax': 0, 'beta1': 0.2, 'beta2': 1.83},
        'blyp': {
            'type': 'pure', 'alpha1': 1.42, 'alpha2': 0.48, 'ax': 0, 'beta1': 0.2, 'beta2': 1.83},
        'bp': {
            'type': 'pure', 'alpha1': 1.42, 'alpha2': 0.48, 'ax': 0, 'beta1': 0.2, 'beta2': 1.83},
        'pbe0': {
            'type': 'hybrid', 'alpha1': 1.42, 'alpha2': 0.48, 'ax': 0.25, 'beta1': 0.2, 'beta2': 1.83},
        'b3lyp': {
            'type': 'hybrid', 'alpha1': 1.42, 'alpha2': 0.48, 'ax': 0.20, 'beta1': 0.2, 'beta2': 1.83},
        'bhlyp': {
            'type': 'hybrid', 'alpha1': 1.42, 'alpha2': 0.48, 'ax': 0.50, 'beta1': 0.2, 'beta2': 1.83},
        'cam-b3lyp': {
            'type': 'rhs', 'alpha1': 1.86, 'alpha2': 0.00, 'ax': 0.38, 'beta1': 0.90, 'beta2': 0},
        'lc-blyp': {
            'type': 'rhs', 'alpha1': 8.0, 'alpha2': 0.00, 'ax': 0.53, 'beta1': 4.50, 'beta2': 0},
        'wb97': {
            'type': 'rhs', 'alpha1': 8.0, 'alpha2': 0.00, 'ax': 0.61, 'beta1': 4.41, 'beta2': 0.0}}
    return d[s]


@overload
def retrieve_hdf5_data(path_hdf5: str | os.PathLike[str], paths_to_prop: str) -> npt.NDArray[Any]:
    ...


@overload
def retrieve_hdf5_data(path_hdf5: str | os.PathLike[str], paths_to_prop: List[str]) -> List[npt.NDArray[Any]]:
    ...


def retrieve_hdf5_data(
    path_hdf5: str | os.PathLike[str],
    paths_to_prop: str | list[str],
) -> npt.NDArray[Any] | list[npt.NDArray[Any]]:
    """Read Numerical properties from ``paths_hdf5``.

    Parameters
    ----------
    path_hdf5
        path to the HDF5
    path_to_prop
        str or list of str to data

    Returns
    -------
    np.ndarray
        array or list of array

    Raises
    ------
    RuntimeError
        The property has not been found

    """
    path_hdf5 = os.fspath(path_hdf5)
    try:
        with h5py.File(path_hdf5, 'r') as f5:
            if isinstance(paths_to_prop, list):
                return [f5[path][()] for path in paths_to_prop]
            else:
                return f5[paths_to_prop][()]
    except KeyError:
        msg = f"There is not {paths_to_prop} stored in the HDF5\n"
        raise KeyError(msg)
    except FileNotFoundError:
        msg = "there is not HDF5 file containing the numerical results"
        raise RuntimeError(msg)


def is_data_in_hdf5(path_hdf5: str | os.PathLike[str], xs: str | List[str]) -> bool:
    """Search if the node exists in the HDF5 file.

    Parameters
    ----------
    path_hdf5
        path to the HDF5
    xs
        either Node path or a list of paths to the stored data

    Returns
    -------
    bool
        Whether the data is stored

    """
    path_hdf5 = os.fspath(path_hdf5)
    if os.path.exists(path_hdf5):
        with h5py.File(path_hdf5, 'r+') as f5:
            if isinstance(xs, list):
                return all(path in f5 for path in xs)
            else:
                return xs in f5
    else:
        return False


def store_arrays_in_hdf5(
    path_hdf5: PathLike,
    paths: str | List[str],
    tensor: np.ndarray | Sequence[np.ndarray],
    dtype: npt.DTypeLike = np.float32,
    attribute: BasisFormats | None = None,
) -> None:
    """Store a tensor in the HDF5.

    Parameters
    ----------
    path_hdf5
        path to the HDF5
    paths
        str or list of nodes where the data is going to be stored
    tensor
        Numpy array or list of array to store
    dtype
        Data type use to store the numerical array
    attribute
        Attribute associated with the tensor

    """
    path_hdf5 = os.fspath(path_hdf5)

    def add_attribute(data_set, k: int = 0):
        if attribute is not None:
            data_set.attrs[attribute.name] = attribute.value[k]

    with h5py.File(path_hdf5, 'r+') as f5:
        if isinstance(paths, list):
            for k, path in enumerate(paths):
                data = tensor[k]
                dset = f5.require_dataset(path, shape=np.shape(data),
                                          data=data, dtype=dtype)
                add_attribute(dset, k)
        else:
            dset = f5.require_dataset(paths, shape=np.shape(
                tensor), data=tensor, dtype=dtype)
            add_attribute(dset)


def change_mol_units(mol: List[AtomXYZ], factor: float = angs2au) -> List[AtomXYZ]:
    """Change the units of the molecular coordinates."""
    new_molecule = []
    for atom in mol:
        coord = tuple(map(lambda x: x * factor, atom.xyz))
        new_molecule.append(AtomXYZ(atom.symbol, coord))  # type: ignore[arg-type]
    return new_molecule


def tuplesXYZ_to_plams(xs: List[AtomXYZ]) -> Molecule:
    """Transform a list of namedTuples to a Plams molecule."""
    plams_mol = Molecule()
    for at in xs:
        symb = at.symbol
        cs = at.xyz
        plams_mol.add_atom(Atom(symbol=symb, coords=tuple(cs)))

    return plams_mol


def number_spherical_functions_per_atom(
    mol: List[AtomXYZ],
    package_name: str,
    basis_name: str,
    path_hdf5: PathLike,
) -> npt.NDArray[np.int_]:
    """Compute the number of spherical shells per atom."""
    ret = []
    with h5py.File(path_hdf5, 'r') as f:
        iterator = ((at_tup[0], valence_electrons[at_tup[0].capitalize()]) for at_tup in mol)
        for atom, q in iterator:
            grp = f[f'{package_name}/basis/{atom}/{basis_name}-q{q}']
            n_funcs = sum(
                calc_n_spherics(grp[f"{i}/coefficients"].attrs['basisFormat'][4:]) for i in grp
            )
            ret.append(n_funcs)
    return np.fromiter(ret, dtype=int)


def calc_n_spherics(fss: npt.NDArray[np.int_]) -> np.int_:
    """Compute the number of spherical shells for a given basis set.

    Most quantum packages use standard basis set which contraction is
    presented usually by a format like:
    c def2-SV(P)
    # c     (7s4p1d) / [3s2p1d]     {511/31/1}
    this mean that this basis set for the Carbon atom uses 7 ``s`` CGF,
    4 ``p`` CGF and 1 ``d`` CGFs that are contracted in 3 groups of 5-1-1
    ``s`` functions, 3-1 ``p`` functions and 1 ``d`` function. Therefore
    the basis set format can be represented by [[5,1,1], [3,1], [1]].

    On the other hand Cp2k uses a special basis set ``MOLOPT`` which
    format explanation can be found at: `C2pk
    <https://github.com/cp2k/cp2k/blob/e392d1509d7623f3ebb6b451dab00d1dceb9a248/cp2k/data/BASIS_MOLOPT>`_.

    Parameters
    ----------
    fss : np.ndarray
        Format basis set

    Returns
    -------
    int
        The number of spherical shells.

    """
    return (fss * np.arange(1, 1 + 2 * len(fss), 2)).sum()


def read_cell_parameters_as_array(
    file_cell_parameters: str | os.PathLike[str],
) -> Tuple[str, npt.NDArray[np.float64]]:
    """Read the cell parameters as a numpy array."""
    arr = np.loadtxt(file_cell_parameters, skiprows=1)

    with open(file_cell_parameters, 'r') as f:
        header = f.readline()

    return header, arr



#! /usr/bin/env python
"""This program plots the energies of each kohn-sham state along the MD trajectory

Note that you have to provide the location of the folder where the NAMD
hamiltonian elements are stored using the -p flag
"""

import numpy as np
import matplotlib.pyplot as plt
import glob
import argparse
import os
from nanoqm.analysis import read_energies


def plot_stuff(energies, ts, ihomo, nhomos, nlumos):
    """
    energies - a vector of energy values that can be plotted
    """
    dim_x = np.arange(ts)

    plt.xlabel('Time (fs)')
    plt.ylabel('Energy (eV)')
    plt.plot(dim_x, energies[:, ihomo - nhomos: ihomo + nlumos])

    fileName = "MOs_energies.png"
    plt.savefig(fileName, format='png', dpi=300)

    plt.show()


def main(path_hams, ts, ihomo, nhomos, nlumos):
    if ts == 'All':
        files = glob.glob(os.path.join(path_hams, 'Ham_*_re'))
        ts = len(files)
    else:
        ts = int(ts)
    energies = read_energies(path_hams, ts)
    plot_stuff(energies, ts, ihomo, nhomos, nlumos)


def read_cmd_line(parser):
    """
    Parse Command line options.
    """
    args = parser.parse_args()

    attributes = ['p', 'ts', 'ihomo', 'nhomos', 'nlumos']

    return [getattr(args, p) for p in attributes]


if __name__ == "__main__":

    msg = "plot_decho -p <path/to/hamiltonians> -ts <time window for analysis>\
    -nhomos <number of homo states to be plotted>\
    -nlumos <number of lumo states to be plotted"

    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument(
        '-p', required=True, help='path to the Hamiltonian files in Pyxaid format')
    parser.add_argument(
        '-ts', type=str, default='All', help='Index of the second state')
    parser.add_argument(
        '-ihomo', type=int, required=True, help='Index of the HOMO state')
    parser.add_argument(
        '-nhomos', default=10, type=int, help='Number of homo states to be plotted')
    parser.add_argument(
        '-nlumos', default=10, type=int, help='Number of lumo states to be plotted')

    main(*read_cmd_line(parser))



"""Common utilities use by the workflows."""

from .. import logger
from ..common import DictConfig, is_data_in_hdf5
from .workflow_single_points import workflow_single_points


def compute_single_point_eigenvalues_coefficients(config: DictConfig):
    """Check if hdf5 contains the required eigenvalues and coefficients.

    If not, it runs the single point calculation.
    """
    node_path_coefficients = f'{config.project_name}/point_0/cp2k/mo/coefficients'
    node_path_eigenvalues = f'{config.project_name}/point_0/cp2k/mo/eigenvalues'

    node_paths = (node_path_coefficients, node_path_eigenvalues)
    if all(is_data_in_hdf5(config.path_hdf5, x) for x in node_paths):
        logger.info("Coefficients and eigenvalues already in hdf5.")
    else:
        # Call the single point workflow to calculate the eigenvalues and
        # coefficients
        logger.info("Starting single point calculation.")
        workflow_single_points(config)



#!/usr/bin/env python

"""Convert old HDF5 files to the new storage layout."""

import argparse
from itertools import chain
from pathlib import Path
from typing import Iterable, Optional, Set

import h5py
import numpy as np


def exists(input_file: str) -> Path:
    """Check if the input file exists."""
    path = Path(input_file)
    if not path.exists():
        raise argparse.ArgumentTypeError(f"{input_file} doesn't exist!")

    return path


class LegacyConverter:
    """Convert legacy HDF5 files to the new storage layout."""

    def __init__(self, source: h5py.File, dest: h5py.File) -> None:
        """Initialize the converter."""
        self.source = source
        self.dest = dest
        self.project = self.get_project_name()

    def get_project_name(self) -> str:
        """Get the project root name."""
        # There are only two names onder root: cp2k and project name
        diff = set(self.source.keys()) - {'cp2k'}
        return diff.pop()

    def copy_data_set(self, old_path: str, new_path: str) -> None:
        """Copy data set from old ``source`` to new ``dest``."""
        if old_path in self.source:
            data = self.source[old_path][()]
            self.dest.require_dataset(new_path, shape=np.shape(data), data=data, dtype=np.float32)

    def copy_node_values(self, old_names: Iterable[str], new_names: Iterable[str]) -> None:
        """Copy the data set values from the old file to the new one."""
        for old, new in zip(old_names, new_names):
            self.copy_data_set(old, new)

    def copy_orbitals(self) -> None:
        """Copy orbitals from old ``source`` to new ``dest``."""
        points = [k for k in self.source[self.project].keys() if k.startswith("point_")]
        keys = {"coefficients", "eigenvalues", "energy"}
        # "project/point_x/cp2k/mo/<coefficients,eigenvalues,energy>"
        old_names = chain(*[[f"{self.project}/{point}/cp2k/mo/{k}" for point in points] for k in keys])
        new_names = chain(*[[f"{k}/{point}" for point in points] for k in keys])
        self.copy_node_values(old_names, new_names)

    def copy_couplings(self) -> None:
        """Copy couplings and swap matrix."""
        couplings = [k for k in self.source[self.project].keys() if k.startswith("coupling_")]
        old_names = [f"{self.project}/{cs}" for cs in couplings]
        new_names = couplings
        self.copy_node_values(old_names, new_names)

        swaps = f"{self.project}/swaps"
        old_names = [swaps]
        new_names = ["swaps"]
        self.copy_node_values(old_names, new_names)

    def copy_overlaps(self) -> None:
        """Copy the overlaps to the new layout."""
        overlaps = [k for k in self.source[self.project].keys() if k.startswith("overlaps_")]
        keys = {"mtx_sji_t0", "mtx_sji_t0_corrected"}
        old_names = chain(*[[f"{self.project}/{over}/{k}" for over in overlaps] for k in keys])
        new_names = chain(*[[f"{over}/{k}" for over in overlaps] for k in keys])
        self.copy_node_values(old_names, new_names)

    def copy_multipoles(self) -> None:
        """Copy the multipoles to the new layout."""
        multipole = f"{self.project}/multipole"
        if multipole in self.source:
            points = [k for k in self.source[multipole].keys() if k.startswith("point_")]
            old_names = [f"{self.project}/multipole/{p}/dipole" for p in points]
            new_names = [f"dipole/{p}" for p in points]
            self.copy_node_values(old_names, new_names)

    def copy_all(self) -> None:
        """Copy all the has been stored in the legacy format."""
        for fun in {"copy_orbitals", "copy_couplings", "copy_overlaps", "copy_multipoles"}:
            method = getattr(self, fun)
            method()


def convert(path_hdf5: Path) -> None:
    """Convert ``path_hdf5`` to new storage format."""
    new_hdf5 = path_hdf5
    old_hdf5 = path_hdf5.rename(f'old_{path_hdf5.name}')

    with h5py.File(new_hdf5, 'a') as dest, h5py.File(old_hdf5, 'r') as source:
        converter = LegacyConverter(source, dest)
        converter.copy_all()


def main():
    """Perform the conversion."""
    parser = argparse.ArgumentParser("convert_legacy_hdf5")
    parser.add_argument("input", type=exists, help="HDF5 file to convert")

    args = parser.parse_args()
    convert(args.input)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""This program plots the average electronic energy during a NAMD simulation averaged over several initial conditions.

It plots both the SH and SE population based energies.

Example:

 plot_cooling.py -p . -nstates 26 -nconds 6

Note that the number of states is the same as given in the pyxaid output.
 It must include the ground state as well.
"""

import argparse

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import interactive

from nanoqm.analysis import (autocorrelate, convolute, read_energies_pyxaid,
                             read_pops_pyxaid, spectral_density)


def func(x, a, b, c, d, e):
    return a * np.exp(- x ** 2 / b ** 2) + d * np.exp(- x / e) + c

def plot_stuff(x_grid, y_grid, x_grid_scaled, y_grid_scaled, sd, w_en, w_en_scaled, nconds, outs, energies, ts, dt):

    plt.figure(1)
    plt.xlabel('Time (fs)')
    plt.ylabel('Energy (eV)')
    for iconds in range(nconds):
        plt.imshow(y_grid[iconds, :, :].T, aspect='auto', extent=(0, len(ts)*dt, np.min(x_grid), np.max(x_grid)), origin='lower', interpolation='bicubic', cmap='hot')
    plt.plot(ts * dt, w_en, 'w')
    interactive(True)
    plt.show()

    plt.figure(2)
    plt.xlabel('Time (fs)')
    plt.ylabel('Excess Energy (eV)')
    for iconds in range(nconds):
        plt.imshow(y_grid_scaled[iconds, :, :].T, aspect='auto', extent=(0, len(ts)*dt, np.min(x_grid_scaled), np.max(x_grid_scaled)), origin='lower', interpolation='bicubic', cmap='hot')  
    plt.plot(ts * dt, w_en_scaled, 'w')
    interactive(True)
    plt.show()

    plt.figure(3) 
    plt.xlabel('Time (fs)')
    plt.ylabel('State Number')
    for iconds in range(nconds):
        plt.imshow(outs[:, :, iconds].T, aspect='auto', origin='lower', extent = ( 0, len(ts) * dt, 0, outs.shape[1]), interpolation='bicubic', cmap='hot')    
    interactive(True)
    plt.show() 

    plt.figure(4)
    plt.xlabel('Energy (eV)')
    plt.ylabel('Counts')
    plt.hist(w_en, bins='auto')
    interactive(True)
    plt.show()

    plt.figure(5)
    plt.xlabel('Energy (eV)')
    plt.ylabel('Counts')
    plt.hist(w_en_scaled, bins='auto')
    interactive(True)
    plt.show()

    plt.figure(6)
    plt.xlabel('Time (fs)')
    plt.ylabel('State Energy (eV)')
    plt.plot(ts*dt, energies[:, :, 0])
    interactive(True)
    plt.show()

    plt.figure(7)
    plt.ylabel('State Number')
    plt.xlabel('Freqencies cm-1')
    sd_int = sd[:, 0, :int(sd.shape[2]/2)] 
    sd_freq = sd[0, 1, :int(sd.shape[2]/2)]
    plt.imshow(sd_int, aspect='auto', origin='lower', extent = (np.min(sd_freq), np.max(sd_freq), 0, sd_int.shape[0] ), interpolation='bicubic', cmap='hot')    
    interactive(False)
    plt.show()


def main(path_output, nstates, nconds, dt, sigma):
    outs = read_pops_pyxaid(path_output, 'out', nstates, nconds)
    energies = read_energies_pyxaid(path_output, 'me_energies', nstates, nconds)
    ##################################
    # Remove the ground state and scale the energies to the lowest excitation energy
    outs = outs[:, 1:, :]
    energies = energies[:, 1:, :]
    lowest_hl_gap = np.amin(energies, axis=1)
    highest_hl_gap = np.amax(energies, axis=1)
    # Convolute a gaussian for each timestep and for each initial condition
    x_grid = np.linspace(np.min(lowest_hl_gap), np.max(highest_hl_gap), 100)
    y_grid = np.stack(np.stack(
        convolute(energies[time, :, iconds], outs[time, :, iconds], x_grid, sigma)
        for time in range(energies.shape[0]) )
        for iconds in range(nconds) )   
    # Scale the energy for computing the excess energy plots
    energies_scaled = np.stack(energies[:, istate, :] - lowest_hl_gap for istate in range(nstates-1)) 
    energies_scaled = energies_scaled.swapaxes(0,1) # Just reshape to have the energies shape (time, nstates, nconds) 
    x_grid_scaled = np.linspace(0, np.max(energies_scaled), 100)
    y_grid_scaled = np.stack(np.stack(
        convolute(energies_scaled[time, :, iconds], outs[time, :, iconds], x_grid_scaled, sigma)
        for time in range(energies_scaled.shape[0]) )
        for iconds in range(nconds) ) 
    # This part is done
    ##################################
    # Compute weighted energies vs population at a given time t
    eav_outs = energies * outs
    el_ene_outs = np.sum(eav_outs, axis=1)
    # Scale them to the lowest excitation energy
    ene_outs_ref0 = el_ene_outs - lowest_hl_gap
    #Average over initial conditions 
    ene_outs_ref0 = np.average(ene_outs_ref0, axis=1) 
    el_ene_av = np.average(el_ene_outs, axis = 1) 
    ##################################
    # Compute autocorrelation function for consecutive pair of states 
    d_en = np.stack(energies[:, istate, 0] - energies[:, istate-1, 0] for istate in range(1, nstates-1))
    acf = np.stack(autocorrelate(d_en[istate, :]) for istate in range(d_en.shape[0]))
    # Compute spectral density
    nacf = acf[:, 1, :]
    sd = np.stack(spectral_density(nacf[istate, :], dt) for istate in range(d_en.shape[0]))
    #################################
    # Call plotting function
    ts = np.arange(energies.shape[0])
    plot_stuff(x_grid, y_grid, x_grid_scaled, y_grid_scaled, sd, el_ene_av, ene_outs_ref0, nconds, outs, energies, ts, dt)

def read_cmd_line(parser):
    """
    Parse Command line options.
    """
    args = parser.parse_args()

    attributes = ['p', 'nstates', 'nconds', 'dt', 'sigma']

    return [getattr(args, p) for p in attributes]


# ============<>===============
if __name__ == "__main__":
    msg = "plot_states_pops -p <path/to/output>\
     -nstates <number of states computed>\
      -nconds <number of initial conditions>\
      -dt <nuclear time step>\
      -sigma <line broadening for convoluton>"
      
    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument('-p', required=True,
                        help='path to the Hamiltonian files in Pyxaid format')
    parser.add_argument('-nstates', type=int, required=True,
                        help='Number of states')
    parser.add_argument('-nconds', type=int, required=True,
                        help='Number of initial conditions')
    parser.add_argument('-dt', type=float, required=False, default=1.0,
                        help='Nuclear Time Step')
    parser.add_argument('-sigma', type=float, required=False, default=0.05,
                        help='Line broadening for the convolution with Gaussian functions')
    main(*read_cmd_line(parser))



#!/usr/bin/env python
"""Command line interface to split a given workflow into several chunks.

Usage:
    distribute_jobs.py -i input.yml

THE USER MUST CHANGES THESE VARIABLES ACCORDING TO HER/HIS NEEDS:
      * project_name
      * path to the basis and Cp2k Potential
      * CP2K:
          - Range of Molecular oribtals printed by CP2K
          - Cell parameter
      * Settings to Run Cp2k simulations
      * Path to the trajectory in XYZ

The slurm configuration is optional but the user can edit it:
    property  default
       * nodes         2
       * tasks        24
       * time   48:00:00
       * name       namd

Otherwise the user can fill the the ``free_format`` property with her
own configuration in the yaml input file.

"""

from __future__ import annotations

import argparse
import os
import shutil
import subprocess
from os.path import join
from typing import Dict, Tuple

import numpy as np
import yaml

from qmflows import Settings

from ..common import DictConfig, read_cell_parameters_as_array, UniqueSafeLoader
from .initialization import split_trajectory
from .input_validation import process_input


def read_cmd_line() -> str:
    """Read the input file and the workflow name from the command line."""
    msg = "distribute_jobs.py -i input.yml"

    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument('-i', required=True,
                        help="Input file in YAML format")

    args = parser.parse_args()
    return args.i


def main() -> None:
    """Distribute the user specified by the user."""
    # command line argument
    input_file = read_cmd_line()

    with open(input_file, 'r') as f:
        args = yaml.load(f, Loader=UniqueSafeLoader)

    # Read and process input
    workflow_type = args['workflow'].lower()
    dict_input = process_input(input_file, workflow_type)
    # Write scripts to run calculations
    if workflow_type == "distribute_derivative_couplings":
        distribute_computations(dict_input, hamiltonians=True)
    else:
        distribute_computations(dict_input)


def distribute_computations(config: DictConfig, hamiltonians: bool = False) -> None:
    """Prepare the computation and write the scripts."""
    # Check if workdir exits otherwise create it
    os.makedirs(config.workdir, exist_ok=True)

    # Split the trajectory in Chunks and move each chunk to its corresponding
    # directory.
    chunks_trajectory = split_trajectory(
        config.path_traj_xyz, config.blocks, config.workdir)
    chunks_trajectory.sort()

    file_cell_parameters = config.cp2k_general_settings.get(
        "file_cell_parameters")
    if file_cell_parameters is not None:
        header_array_cell_parameters = read_cell_parameters_as_array(
            file_cell_parameters)

    # Scratch for all the chunks
    parent_scratch = config.scratch_path
    accumulated_number_of_geometries = 0

    for index, file_xyz in enumerate(chunks_trajectory):
        copy_config = DictConfig(config.copy())
        copy_config["scratch_path"] = os.path.join(parent_scratch, f"scratch_chunk_{index}")

        folder_path = os.path.abspath(join(copy_config.workdir, f'chunk_{index}'))

        dict_input = DictConfig({
            'folder_path': folder_path, "file_xyz": file_xyz, 'index': index})

        create_folders(copy_config, dict_input)

        # number of geometries per batch
        dim_batch = compute_number_of_geometries(join(folder_path, file_xyz))

        # change the window of molecules to compute
        copy_config['enumerate_from'] = accumulated_number_of_geometries

        # HDF5 file where both the Molecular orbitals and coupling are stored
        copy_config.path_hdf5 = join(copy_config.scratch_path, f'chunk_{index}.hdf5')

        # Change hdf5 and trajectory path of each batch
        copy_config["path_traj_xyz"] = file_xyz

        if file_cell_parameters is not None:
            add_chunk_cell_parameters(
                header_array_cell_parameters, copy_config, dict_input)

        # files with PYXAID
        if hamiltonians:
            path_ham = "hamiltonians" if not config.orbitals_type else f"{config.orbitals_type}_hamiltonians"
            dict_input.hamiltonians_dir = join(
                copy_config.scratch_path, path_ham)

        # Write input file
        write_input(folder_path, copy_config)

        # Slurm executable
        scheduler = copy_config.job_scheduler["scheduler"].upper()
        if scheduler == "SLURM":
            write_slurm_script(copy_config, dict_input, dim_batch, accumulated_number_of_geometries)
        else:
            msg = f"The request job_scheduler: {scheduler} it is not implemented"
            raise RuntimeError(msg)

        accumulated_number_of_geometries += dim_batch


def write_input(folder_path: str | os.PathLike[str], original_config: DictConfig) -> None:
    """Write the python script to compute the PYXAID hamiltonians."""
    file_path = join(folder_path, "input.yml")

    # transform settings to standard dictionary
    config = Settings(original_config).as_dict()

    # basis and potential
    config["cp2k_general_settings"]["path_basis"] = os.path.abspath(
        config["cp2k_general_settings"]["path_basis"])

    # remove unused keys from input
    for k in ['blocks', 'job_scheduler', 'mo_index_range',
              'workdir']:
        del config[k]

    # rename the workflow to execute
    dict_distribute = {"distribute_derivative_couplings": "derivative_couplings",
                       "distribute_absorption_spectrum": "absorption_spectrum",
                       "distribute_single_points": "single_points"
                       }
    workflow_type = config["workflow"].lower()
    config['workflow'] = dict_distribute[workflow_type]
    with open(file_path, "w") as f:
        yaml.dump(config, f, default_flow_style=False, allow_unicode=True)


def create_folders(config: DictConfig, dict_input: DictConfig) -> None:
    """Create folder for each batch and copy the xyz."""
    # Move xyz to temporal file
    os.makedirs(dict_input.folder_path, exist_ok=True)
    shutil.move(dict_input.file_xyz, dict_input.folder_path)

    # Scratch directory
    batch_dir = join(config.scratch_path, f'batch_{dict_input.index}')
    os.makedirs(batch_dir, exist_ok=True)


def write_slurm_script(config: DictConfig, dict_input: DictConfig,
                       dim_batch: int, acc: int) -> None:
    """Write an Slurm launch script."""
    index = dict_input.index
    python = "\n\nrun_workflow.py -i input.yml\n"
    results_dir = "results_chunk_" + str(index)
    mkdir = f"\nmkdir -p {results_dir}\n"
    slurm_config = config.job_scheduler

    # Copy a subset of Hamiltonians
    if dict_input.get("hamiltonians_dir") is None:
        copy = ""
    else:
        range_batch = (acc, acc + dim_batch - 1)
        files_hams = f"{dict_input.hamiltonians_dir}/Ham_{{{range_batch[0]}..{range_batch[1]}}}_*"
        copy = f'cp -r {config.path_hdf5} {files_hams} {results_dir}\n'

    # Script content
    content = format_slurm_parameters(slurm_config) + python + mkdir + copy

    # Write the script
    with open(join(dict_input.folder_path, "launch.sh"), 'w') as f:
        f.write(content)


def format_slurm_parameters(slurm: Dict[str, str]) -> str:
    """Format as a string some SLURM parameters."""
    sbatch = "#SBATCH -{} {}\n".format

    header = "#! /bin/bash\n"
    time = sbatch('t', slurm["wall_time"])
    nodes = sbatch('N', slurm["nodes"])
    tasks = sbatch('n', slurm["tasks"])
    name = sbatch('J', slurm["job_name"])
    queue = sbatch('p', slurm["queue_name"])

    modules = slurm["load_modules"]

    if "free_format" in slurm:
        # Remove empty spaces
        lines = slurm["free_format"].splitlines()
        return '\n'.join(' '.join(x.split()) for x in lines if x)
    else:
        return ''.join((header, time, nodes, tasks, name, queue, modules))


def compute_number_of_geometries(file_name: str | os.PathLike[str]) -> int:
    """Count the number of geometries in XYZ formant in a given file."""
    with open(file_name, 'r') as f:
        numat = int(f.readline())

    cmd = f"wc -l {os.fspath(file_name)}"
    wc = subprocess.getoutput(cmd).split()[0]

    lines_per_geometry = numat + 2

    return int(wc) // lines_per_geometry


def add_chunk_cell_parameters(
        header_array_cell_parameters: Tuple[str, np.ndarray],
        config: DictConfig, dict_input: DictConfig) -> None:
    """Add the corresponding set of cell parameters for a given chunk."""
    path_file_cell_parameters = join(
        dict_input.folder_path, "cell_parameters.txt")

    # Adjust settings
    config.cp2k_general_settings["file_cell_parameters"] = path_file_cell_parameters

    # extract cell parameters for a chunk
    header, arr = header_array_cell_parameters
    size = int(np.ceil(arr.shape[0] / config.blocks))
    low = dict_input.index * size
    high = low + size
    step = config.stride
    if high < arr.shape[0]:
        matrix_cell_parameters = arr[low:high:step, :]
    else:
        matrix_cell_parameters = arr[low::step, :]

    # adjust indices of the cell parameters
    size = matrix_cell_parameters.shape[0]
    matrix_cell_parameters[:, 0] = np.arange(size)
    # Save file
    fmt = '%d %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f'
    np.savetxt(path_file_cell_parameters,
               matrix_cell_parameters, header=header, fmt=fmt)


if __name__ == "__main__":
    main()



"""Simulation workflows."""
from .initialization import initialize
from .workflow_coop import workflow_crystal_orbital_overlap_population
from .workflow_coupling import workflow_derivative_couplings
from .workflow_single_points import workflow_single_points
from .workflow_stddft_spectrum import workflow_stddft

__all__ = [
    'initialize', 'workflow_crystal_orbital_overlap_population',
    'workflow_derivative_couplings', 'workflow_single_points', 'workflow_stddft']



"""Installation recipe."""
import os
import sys
from os.path import join
from typing import TYPE_CHECKING

import setuptools
import pybind11
from Cython.Distutils import build_ext
from setuptools import Extension, find_packages, setup

if TYPE_CHECKING:
    from distutils.ccompiler import CCompiler
    from distutils.errors import CompileError
else:
    CCompiler = setuptools.distutils.ccompiler.CCompiler
    CompileError = setuptools.distutils.errors.CompileError

here = os.path.abspath(os.path.dirname(__file__))

version: "dict[str, str]" = {}
with open(os.path.join(here, 'nanoqm', '_version.py'), 'r', encoding='utf8') as f:
    exec(f.read(), version)


def readme() -> str:
    """Load readme."""
    with open('README.rst', 'r', encoding='utf8') as f:
        return f.read()


def has_flag(compiler: "CCompiler", flagname: str) -> bool:
    """Return a boolean indicating whether a flag name is supported on the specified compiler.

    As of Python 3.6, CCompiler has a `has_flag` method.
    http: // bugs.python.org/issue26689
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cc') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except CompileError:
            return False
    return True


def cpp_flag(compiler: "CCompiler") -> str:
    """Return the -std=c++[17/14/11] compiler flag.

    The newer version is prefered over c++11 (when it is available).
    """
    flags = ['-std=c++17', '-std=c++14', '-std=c++11']
    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError('Unsupported compiler -- at least C++11 support '
                       'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    c_opts: "dict[str, list[str]]" = {
        'msvc': ['/EHsc'],
        'unix': ['-Wall'],
    }
    l_opts: "dict[str, list[str]]" = {
        'msvc': [],
        'unix': ['-Wall'],
    }

    if sys.platform == 'darwin':
        darwin_opts = ['-stdlib=libc++', '-mmacosx-version-min=10.14', '-fno-sized-deallocation']
        c_opts['unix'] += darwin_opts
        l_opts['unix'] += darwin_opts

    def build_extensions(self) -> None:
        """Actual compilation."""
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts
        build_ext.build_extensions(self)


def parse_requirements(path: "str | os.PathLike[str]") -> "list[str]":
    """Parse a ``requirements.txt`` file and strip all empty and commented lines."""
    ret = []
    with open(path, "r", encoding="utf8") as f:
        for i in f:
            j = i.split("#", 1)[0].strip().rstrip()
            if j:
                ret.append(j)
    return ret


def get_paths() -> "tuple[list[str], list[str]]":
    """Get the paths specified in the ``QMFLOWS_INCLUDEDIR`` and ``QMFLOWS_LIBDIR`` \
    environment variables.

    If not specified if specified use ``CONDA_PREFIX`` instead.
    Multiple include and/or lib paths must be specified with the standard (OS-specific)
    path separator, *e.g.* ``":"`` for POSIX.

    Examples
    --------
    .. code-block:: bash
        export QMFLOWS_INCLUDEDIR="/libint/include:/eigen3/include"
        export QMFLOWS_LIBDIR="/hdf5/lib:/libint/lib"

    .. code-block:: python
        >>> get_paths()
        (['/libint/include', '/eigen3/include'], ['/hdf5/lib', '/libint/lib'])

    Returns
    -------
    tuple[list[str], list[str]]
        Lists of include- and library-directories used in compiling the ``compute_integrals``
        extension module.

    """
    conda_prefix = os.environ.get("CONDA_PREFIX")
    include_dirs = os.environ.get("QMFLOWS_INCLUDEDIR")
    lib_dirs = os.environ.get("QMFLOWS_LIBDIR")

    if include_dirs is not None and lib_dirs is not None:
        include_list = include_dirs.split(os.pathsep) if include_dirs else []
        lib_list = lib_dirs.split(os.pathsep) if lib_dirs else []
    elif conda_prefix is not None:
        include_list = [
            join(conda_prefix, "include"),
            join(conda_prefix, "include", "eigen3"),
        ]
        lib_list = [join(conda_prefix, "lib")]
    else:
        raise RuntimeError(
            "No conda module found. A Conda environment is required "
            "or one must set both the `QMFLOWS_INCLUDEDIR` and `QMFLOWS_LIBDIR` "
            "environment variables"
        )
    return include_list, lib_list


include_list, lib_list = get_paths()
ext_pybind = Extension(
    'nanoqm.compute_integrals',
    sources=['libint/compute_integrals.cc'],
    include_dirs=[
        "libint/include",
        *include_list,
        pybind11.get_include(),
    ],
    libraries=['hdf5', 'int2'],
    library_dirs=lib_list,
    language='c++',
)

setup(
    name='nano-qmflows',
    version=version['__version__'],
    description='Derivative coupling calculation',
    license='Apache-2.0',
    url='https://github.com/SCM-NV/nano-qmflows',
    author='Felipe Zapata & Ivan Infante',
    author_email='f.zapata@esciencecenter.nl',
    keywords='chemistry Photochemistry Simulation',
    long_description=readme() + '\n\n',
    long_description_content_type='text/x-rst',
    packages=find_packages(),
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Typing :: Typed',
    ],
    install_requires=parse_requirements("install_requirements.txt"),
    cmdclass={'build_ext': BuildExt},
    python_requires='>=3.7',
    ext_modules=[ext_pybind],
    extras_require={
        'test': parse_requirements("test_requirements.txt"),
        'doc': parse_requirements("doc_requirements.txt"),
    },
    include_package_data=True,
    package_data={
        'nanoqm': ['basis/*.json', 'basis/BASIS*', 'basis/GTH_POTENTIALS', 'py.typed', '*.pyi']
    },
    entry_points={
        'console_scripts': [
            'run_workflow.py=nanoqm.workflows.run_workflow:main',
            'distribute_jobs.py=nanoqm.workflows.distribute_jobs:main'
        ]
    },
    scripts=[
        'scripts/convert_legacy_hdf5.py',
        'scripts/hamiltonians/plot_mos_energies.py',
        'scripts/hamiltonians/plot_spectra.py',
        'scripts/pyxaid/plot_average_energy.py',
        'scripts/pyxaid/plot_cooling.py',
        'scripts/pyxaid/plot_spectra_pyxaid.py',
        'scripts/pyxaid/plot_states_pops.py',
        'scripts/qmflows/mergeHDF5.py',
        'scripts/qmflows/plot_dos.py',
        'scripts/qmflows/rdf.py',
        'scripts/qmflows/removeHDF5folders.py',
        'scripts/qmflows/remove_mos_hdf5.py',
        'scripts/qmflows/convolution.py'
    ]
)



"""Nonadiabatic coupling implementation."""
from .nonAdiabaticCoupling import (calculate_couplings_3points,
                                   calculate_couplings_levine,
                                   compute_overlaps_for_coupling,
                                   correct_phases)

__all__ = ['calculate_couplings_3points', 'calculate_couplings_levine',
           'calculate_couplings_levine', 'compute_overlaps_for_coupling',
           'correct_phases']



#! /usr/bin/env python
""" This script convolutes the calculated oscillator strengts for different snapshots in a MD trajectory and plot the average spectrum.

Usage:
convolution.py -sigma 0.05 -n 2 -nm True -write True
Use the sigma flag to change the sigma parameter of the gaussian functions
used in the convolution. If you use the n flag you imply that you want to plot
only the spectrum of snapshot number n (starting from 0)instead of an average
over all strucutures.
Use the nm to obtain a plot in nm scale.
You can choose to write the coordinates obtained from the convolution with the write flag.
"""

import argparse
import glob

import matplotlib.pyplot as plt
import numpy as np

from nanoqm.analysis import convolute


def main(sigma, n, nm, write):
    # Check output files
    files = sorted(glob.glob('output_*.txt'))

    if n is None:
        # Define energy window for the plot
        energy = np.loadtxt(files[0], usecols=1)
        emax = energy[-1] + 0.5
        emin = energy[0] - 0.5
        x_grid = np.linspace(emin, emax, 800)
        y_grid = np.empty([x_grid.size, 0])
        for f in files:
            # Read transition energies and oscillator strengths
            data = np.loadtxt(f, usecols=(1, 2))
            # Convolute each spectrum at a time
            ys = convolute(data[:, 0], data[:, 1], x_grid, sigma)
            # Stack the spectra in a matrix
            y_grid = np.column_stack((y_grid, ys))
        # Average them
        y_grid = np.sum(y_grid, axis=1) / len(files)

    else:
        # Read transition energies and oscillator strengths
        data = np.loadtxt(files[n], usecols=(1, 2))
        # Define energy window for the plot
        emax = data[-1, 0] + 0.5
        emin = data[0, 0] - 0.5
        x_grid = np.linspace(emin, emax, 800)
        # Convolute
        y_grid = convolute(data[:, 0], data[:, 1], x_grid, sigma)

    # Convert in nm if request
    if nm:
        x_grid = 1240/x_grid

    # plot
    plt.plot(x_grid, y_grid)
    if nm:
        plt.xlabel('Wavelenght[nm]')
    else:
        plt.xlabel('Energy[eV]')
    plt.ylabel('Oscillator strength')
    plt.show()

    # Write convoluted coordinates if request
    if write:
        output = np.empty((len(x_grid), 2))
        output[:, 0] = x_grid
        output[:, 1] = y_grid
        fmt = '{:^10s}{:^10s}'
        if nm:
            header = fmt.format(
                'nm', 'f')
            output = output[::-1]
        else:
            header = fmt.format(
                'eV', 'f')
        np.savetxt('convolution.txt', output,
                   fmt='%11.3f %10.5f', header=header)


def read_cmd_line(parser):
    """
    Parse Command line options.
    """
    args = parser.parse_args()

    attributes = ['sigma', 'n', 'nm', 'write']

    return [getattr(args, p) for p in attributes]


# ============<>===============
if __name__ == "__main__":

    msg = """convolution.py -sigma <sigma parameter of the gaussian functions> -n <plot only the spectrum of structure n>
    -nm <plot in nm> -write <write the coordinates from the convolution>"""

    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument('-sigma', default=0.1, type=float,
                        help='Sigma parameter of the gaussian functions')
    parser.add_argument('-n', default=None, type=int,
                        help='Plot only the spectrum of the strucure number n')
    parser.add_argument('-nm', default=False, type=bool,
                        help='Convert the x axis in nm')
    parser.add_argument('-write', default=False, type=bool,
                        help='Write the coordinates from the convolution')

    main(*read_cmd_line(parser))



#! /usr/bin/env python
"""This program reads the ouput files, out and me_pop,from a NAMD simulation run with pyxaid.

It average the populations of each state upon several initial conditions and
also allows to define macrostates.
A macrostate is formed by a group of (micro)states indexed according to pyxaid numbering.
For example, if you have donor-acceptor system, you can group all (micro)states formed by
excited states localized within the donor/acceptor and or charge transfer states
into a few macrostates. This helps the analysis of the population traces when a large
number of excited states is included in the NAMD simulation.

Example:

 plot_states_pops.py -p . -ms "[ [0], [1,3], [4,6] ]" -nconds 6

This means that the path with the output files is the current path '.' .
The number of initial conditions over which the populations are averaged are 6.
The macrostates are defined as a list of lists. The [0] is the ground state, [1,3]
form a macrostate of states 1 and 3 indexed as in pyxaid, [4,6] forms
a macrostates of states 4 and 6.
"""

import numpy as np
import os
import matplotlib.pyplot as plt
import argparse
from nanoqm.analysis import parse_list_of_lists


def plot_stuff(outs, pops):
    """energies - a vector of energy values that can be plotted."""
    dim_x = np.arange(outs.shape[0])

    ax1 = plt.subplot(121)
    ax1.set_title('SH Population ')
    ax1.set_xlabel('Time (fs)')
    ax1.set_ylabel('Population')
    ax1.plot(dim_x, outs[0:, :])

    ax2 = plt.subplot(122)
    ax2.set_title('SE Population ')
    ax2.set_xlabel('Time (fs)')
    ax2.set_ylabel('Population')
    ax2.plot(dim_x, pops[0:, :])

    fileName = "State Population.png"
    plt.savefig(fileName, format='png', dpi=300)

    plt.show()


def read_populations(path, fn, nconds, ms):
    inpfile = os.path.join(path, fn)
    cols = list(map(lambda row: tuple(map(lambda x: x * 2 + 3, row)), ms))
    xs = [np.stack(np.loadtxt(f'{inpfile}{j}', usecols=col)
                   for j in range(nconds)) for col in cols]
    return xs


def main(path_output, ms, nconds):

    outs = read_populations(path_output, 'out', nconds, ms)
    pops = read_populations(path_output, 'me_pop', nconds, ms)

    outs_avg = [np.average(out, axis=0) for out in outs]
    pops_avg = [np.average(pop, axis=0) for pop in pops]

    outs_fin, pops_fin = [], []

    for x in outs_avg:
        if x.ndim == 1:
            outs_fin.append(x)
        else:
            outs_fin.append(np.sum(x, axis=1))
    outs_fin = np.array(outs_fin).transpose()

    for x in pops_avg:
        if x.ndim == 1:
            pops_fin.append(x)
        else:
            pops_fin.append(np.sum(x, axis=1))
    pops_fin = np.array(pops_fin).transpose()

    plot_stuff(outs_fin, pops_fin)


def read_cmd_line(parser):
    """
    Parse Command line options.
    """
    args = parser.parse_args()

    attributes = ['p', 'ms', 'nconds']

    return [getattr(args, p) for p in attributes]


# ============<>===============
if __name__ == "__main__":
    msg = "plot_states_pops -p <path/to/output>\
    -ms <list of microstates state indexes>\
    -nconds <number of initial conditions>"

    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument('-p', required=True,
                        help='path to the Hamiltonian files in Pyxaid format')
    parser.add_argument('-ms', type=str, required=True,
                        help='Macrostate defined as a list of microstates')
    parser.add_argument('-nconds', type=int, required=True,
                        help='Number of initial conditions')

    p, ms, nconds = read_cmd_line(parser)
    main(p, parse_list_of_lists(ms), nconds)
